#!/usr/bin/perl


####    Settings

$cfgpoi     = "poi.cfg";
$cfgpoly    = "poly.cfg";
$cfgheader  = "header.cfg";


$codepage = "1251";


####    Action


use POSIX qw(strftime);

print STDERR "\n  ---|   OSM -> MP converter  v0.4    (c) 2008  liosha, xliosha\@gmail.com\n\n";



####    Reading configs

open CFG, $cfgpoi;
while (<CFG>) { 
   if ( (!$_) || /^ *[\#\;]/ ) { next; }
   chomp;
   ($k, $v, $type, $llev, $hlev, @p) = split / +/;
   if ($type) {
     $poitype{"$k=$v"} = $type;
     $poillev{"$k=$v"} = ( ($llev ne "") ? $llev : 0 );
     $poihlev{"$k=$v"} = ( ($hlev ne "") ? $hlev : 1 );
     $poirest{"$k=$v"} = join " ", @p;  
   }
}
close CFG;


open CFG, $cfgpoly;
while (<CFG>) { 
   if ( (!$_) || /^ *[\#\;]/ ) { next; }
   chomp;
   ($k, $v, $mode, $type, $llev, $hlev, $rp, @p) = split / +/;
   if ($type) {
     $polymode{"$k=$v"} = $mode;
     $polytype{"$k=$v"} = $type;
     $polyllev{"$k=$v"} = ( ($llev ne "") ? $llev : 0 );
     $polyhlev{"$k=$v"} = ( ($hlev ne "") ? $hlev : 0 );
     $polyrout{"$k=$v"} = $rp           if ($mode eq "r");
     $polyrest{"$k=$v"} = join " ", @p;
   }
}
close CFG;



####    Header

open HEAD, $cfgheader;
print <HEAD>;

open IN, $ARGV[0];


print "\n; Converted from OpenStreetMap data  " . strftime ("%Y-%m-%d %H:%M:%S", localtime) . "\n\n";


####    Loading nodes and writing POIs

print "\n\n\n; ### Points\n\n";

while (<IN>) {             
   last if /\<way/;

   if ( /\<node/ ) {                   
      /^.*id=["'](\-?\d+)["'].*lat=["'](\-?\d+\.?\d*)["'].*lon=["'](\-?\d+\.?\d*)["'].*$/;
      $id = $1;
      $latlon = "$2,$3";
      $nodes{$1} = $latlon;
      undef ($poi);
      undef ($poiname);
      next;
   }

   if ( /\<tag/ ) {                   
      /^.*k=["'](.*)["'].*v=["'](.*)["'].*$/;
      $poi = "$1=$2"                            if ($poitype{"$1=$2"});
      $poiname = convert_string ($2)            if ($1 eq "name");
      next;
   }

   if ( /\<\/node/ ) {                   

      if ($poi) {
         print  "; NodeID = $id\n";
         print  "; $poi\n";
         printf "[POI]\nType=%s\nData%d=($latlon)\n", $poitype{$poi}, $poillev{$poi};
         printf "EndLevel=%d\n", $poihlev{$poi}         if ($poihlev{$poi} > $poillev{$poi});
         print  "Label=$poiname\n"                      if ($poiname);
         printf "%s\n", $poirest{$poi}                  if ($poirest{$poi});
         print  "[END]\n\n";
      }
   }
}


$waypos  = tell IN;
$waystr  = $_;



####    Ways, 1st pass: detect road nodes

while ($_) {
   
   last if /\<relation/;

   if ( /\<way/ ) {                   
      /^.*id=["'](\-?\d+)["'].*$/;
      $id = $1;
      @chain = ();
      undef ($poly);
      next;
   }

   if ( /\<nd/ ) {                   
      /^.*ref=["'](.*)["'].*$/;
      if ($nodes{$1}  &&  $1 ne $chain[$#chain] ) {
          push @chain, $1;
      }
      next;
   }

   if ( /\<tag/ ) {                   
      /^.*k=["'](.*)["'].*v=["'](.*)["'].*$/;
      $poly = "$1=$2"  if ($polytype{"$1=$2"});

      next;
   }

   if ( /\<\/way/ ) {                   

      if ($polymode{$poly} eq "r") {
          $bnodes{$chain[0]} ++ ;
          $bnodes{$chain[$#chain]} ++ ;
          for $node (@chain) {
              $rnodes{$node} ++ ;
          }
      }
   }

} continue { $_ = <IN>; }


$relpos  = tell IN;
$relstr  = $_;


####    Relations, 1st pass

while ($_) {

    if ( /\<relation/ ) {                   
        /^.*id=["'](\-?\d+)["'].*$/;
        $id = $1;

        undef $outer;
        undef @inner;
        undef $reltype;
        next;
    }

    if ( /\<member/ ) {                   
        /type=["'](\w+)["'].*ref=["'](\-?\d+)["'].*role=["'](\w+)["']/;

        $outer = $2             if ($3 eq "outer" && $1 eq "way");
        push @inner, $2         if ($3 eq "inner" && $1 eq "way");

        next;
    }

    if ( /\<tag/ ) {
        /k=["'](\w+)["'].*v=["'](\w+)["']/;

        $reltype = $2           if ( $1 eq "type" );

        next;
    }

    if ( /\<\/relation/ ) {                   

        if ( $reltype eq "multipolygon" ) {

            $multiout{$outer} = "-";
            @multiinn{@inner} = @inner;
        }
        
        next;
    }

} continue { $_ = <IN>; }




####    Dumping NodID list


print "\n\n\n; ### Routing nodes\n\n";

$nodenum = 1;
while ($node = each %rnodes) {
    if ($rnodes{$node} > 1 || $bnodes{$node} > 0) {
        printf "; NodID=$nodenum:   ID=$node, ${rnodes{$node}} ways%s\n", (${bnodes{$node}} ? "  ${bnodes{$node}} ends" : "");

        $nodid{$node} = $nodenum;
        $nodenum ++ ;
    } 
}




%yesno = (  "yes"       => 1,
            "true"      => 1,
            "1"         => 1,
            "no"        => 0,
            "false"     => 0,
            "0"         => 0);  


####    Ways, 2nd pass

print "\n\n\n; ### Polys\n\n";

seek IN, $waypos, 0;
$_ = $waystr;

$roadcount = 1;

while ($_) {
   
   last if /\<relation/;

   if ( /\<way/ ) {                   
      /^.*id=["'](\-?\d+)["'].*$/;
      $id = $1;

      $dupcount = 0;
      @chain = ();
      @breaks = (0);

      $nodcount = 0;
      %nodelist = ();

      undef ($poly);
      undef ($polyname);
      undef ($polyname2);

      undef ($polydir);
      undef ($polytoll);
      undef ($polynoauto);
      undef ($polynobus);
      undef ($polynoped);
      undef ($polynobic);
      undef ($polynohgv);

      next;
   }

   if ( /\<nd/ ) {                   
      /^.*ref=["'](.*)["'].*$/;
      if ($nodes{$1}) {

          if ( $1 == $chain[$#chain] )  {   
              $dupcount ++;           
          } else {
              push @breaks, $nodcount-1
                  if ( scalar (grep { $_ eq $1 } @chain) > 0 );

              push @chain, $1;        
              $nodelist{$nodcount} = $1     if ($nodid{$1});
              $nodcount ++ ;
          }      

      }
      next;
   }

   if ( /\<tag/ ) {                   
      /^.*k=["'](.*)["'].*v=["'](.*)["'].*$/;

      $poly = "$1=$2"  if ($polytype{"$1=$2"});
      $polyname = convert_string ($2)  if ($1 eq "name");
      $polyname2 = convert_string ($2)  if ($1 eq "ref");

      $polydir     = $yesno{$2}         if ($1 eq "oneway");
      $polytoll    = $yesno{$2}         if ($1 eq "toll");
      $polynoauto  = 1 - $yesno{$2}     if ($1 eq "motorcar");
      $polynobus   = 1 - $yesno{$2}     if ($1 eq "psv");
      $polynoped   = 1 - $yesno{$2}     if ($1 eq "foot");
      $polynobic   = 1 - $yesno{$2}     if ($1 eq "bicycle");
      $polynohgv   = 1 - $yesno{$2}     if ($1 eq "hgv");
                        
      next;
   }


   if ( /\<\/way/ && $multiinn{$id} ) {
       $multiinn{$id} = join ("), (", @nodes{@chain});
   }


   if ( /\<\/way/ && $poly ) {

      print  "; WayID = $id\n";
      print  "; $poly\n";

      print  "; ERROR: WayID=$id has dupes\n"
          if ( $dupcount > 0 );

      if ( $polymode{$poly} eq "r" ) {

         $d = $dr = "";
         
         if (scalar keys %nodelist < 2) {
              print "; ERROR: WayID=$id road has too few nodes\n";
              $d = "; ";
         }

         push @breaks, $nodcount-1;
         if (scalar @breaks > 2) {
             printf "; ERROR: WayID=$id road has self-intersections\n"
                 if ( scalar @breaks > 3  ||  $chain[0] != $chain[$#chain] );
             printf "; FIXED: this road is split at %s\n", join (", ", @breaks[1 .. $#breaks-1]);
         }

         @rp = split /,/, $polyrout{$poly};
         $rp[2]  = $polydir                         if defined $polydir;
         $rp[3]  = $polytoll                        if defined $polytoll;
         $rp[5]  = $rp[6] = $rp[8] = $polynoauto    if defined $polynoauto;
         $rp[7]  = $polynobus                       if defined $polynobus;
         $rp[9]  = $polynoped                       if defined $polynoped;
         $rp[10] = $polynobic                       if defined $polynobic;
         $rp[11] = $polynohgv                       if defined $polynohgv;

         for ( $i=0; $i<$#breaks; $i++ ) {
             print "\n; Slice from ${breaks[$i]} to ${breaks[$i+1]}\n"
                 if ( $#breaks > 1 );
          
             if ( ! defined $nodid{$chain[$breaks[$i+1]]} ) {
                 print "; Added NodID=$nodenum at ${breaks[$i+1]}\n";
                 $nodid{$chain[$breaks[$i+1]]} = $nodenum;
                 $nodenum ++ ;
                 $nodelist{$breaks[$i+1]} = $chain[$breaks[$i+1]];
             }

             print  "${d}[POLYLINE]\n";
             printf "${d}Type=%s\n",        $polytype{$poly};
             printf "${d}EndLevel=%d\n",    $polyhlev{$poly}        if ($polyhlev{$poly} > $polyllev{$poly});
             print  "${d}Label=$polyname\n"                         if ($polyname);
             print  "${d}Label2=~[0x04]$polyname2\n"                if ($polyname2);
             print  "${d}DirIndicator=$polydir\n"                   if defined $polydir;

             printf "${d}Data%d=(%s)\n", $polyllev{$poly}, join ("), (", @nodes{@chain[$breaks[$i] .. $breaks[$i+1]]});
             print  "${d}${dr}RoadID=$roadcount\n";
             print  "${d}${dr}RouteParams=" . join (",", @rp) . "\n";

             $roadnod = 1;
             for $nod (sort {$a <=> $b} keys %nodelist) {
                 if ($nod >= $breaks[$i] && $nod <= $breaks[$i+1]) {
                     printf "${d}${dr}Nod$roadnod=%d,${nodid{$nodelist{$nod}}}\n", $nod-$breaks[$i] ;
                     $roadnod ++ ;
                 }
             }
             $roadcount ++ ;
             
             print  "${d}[END]\n";
         }

         print  "\n\n";

         next;
      } 
      

      if ( $polymode{$poly} eq "l" ) {

         $d = "";

         if ( scalar @chain < 2 ) {
             print  "; ERROR: WayID=$id polyline has too few nodes\n";
             $d = "; ";
         }

         print  "${d}[POLYLINE]\n";
         printf "${d}Type=%s\n",        $polytype{$poly};
         printf "${d}Data%d=(%s)\n",    $polyllev{$poly}, join ("), (", @nodes{@chain});
         printf "${d}EndLevel=%d\n",    $polyhlev{$poly}        if ($polyhlev{$poly} > $polyllev{$poly});
         print  "${d}Label=$polyname\n"                         if ($polyname);
         print  "${d}DirIndicator=$polydir\n"                   if defined $polydir;
         print  "${d}[END]\n\n\n";

         next;
      }
      

      if ( $polymode{$poly} eq "p" ) {

         $d = "";

         if ( scalar @chain < 4 ) {
             print  "; ERROR: WayID=$id area has too few nodes\n";
             $d = "; ";
         }
         if ( $chain[0] ne $chain[$#chain] ) {
             print  "; ERROR: WayID=$id area is not closed\n";
         }

         my $polytext;
         $polytext  = sprintf "${d}Type=%s\n",        $polytype{$poly};
         $polytext .= sprintf "${d}EndLevel=%d\n",    $polyhlev{$poly}        if ($polyhlev{$poly} > $polyllev{$poly});
         $polytext .= sprintf "${d}Label=$polyname\n"                         if ($polyname);
         $polytext .= sprintf "${d}Data%d=(%s)\n",    $polyllev{$poly}, join ("), (", @nodes{@chain});


         if ( $multiout{$id} ) {
             print "; multipolygon, see later\n\n\n";
             $multiout{$id} = $polytext;
         } else {
             print  "${d}[POLYGON]\n";
             print  $polytext;
             print  "${d}[END]\n\n\n";
         }

         next;
      }
   }
} continue { $_ = <IN>; }




####    Relations, 2nd pass

print "\n; ####    Multipolygons\n\n\n";

while ($_) {

    if ( /\<relation/ ) {                   
        /^.*id=["'](\-?\d+)["'].*$/;
        $id = $1;

        undef $outer;
        undef @inner;
        undef $reltype;
        next;
    }

    if ( /\<member/ ) {                   
        /type=["'](\w+)["'].*ref=["'](\-?\d+)["'].*role=["'](\w+)["']/;

        $outer = $2             if ($3 eq "outer" && $1 eq "way");
        push @inner, $2         if ($3 eq "inner" && $1 eq "way");

        next;
    }

    if ( /\<tag/ ) {
        /k=["'](\w+)["'].*v=["'](\w+)["']/;

        $reltype = $2           if ( $1 eq "type" );

        next;
    }

    if ( /\<\/relation/ ) {                   

        if ( $reltype eq "multipolygon" && $multiout{$outer} ne "-" ) {

            print "; WayID=$outer - multipolygon RelationID=$id\n";
            print "[POLYGON]\n";
            print $multiout{$outer};
            $multiout{$outer} =~ /Data(\d)=/;
            for $hole ( @inner ) {
               print "; WayID=$hole - hole\n";
               print "Data$1=(${multiinn{$hole}})\n"  if ($multiinn{$hole} ne $hole);
            };
            print "[END]\n\n";
        }
        
        next;
    }

} continue { $_ = <IN>; }



####    Functions

use Encode;

sub convert_string {

   my $str = encode ("cp".$codepage, decode("utf8", $_[0]));

   $str =~ s/\&#38\;/\&/gi;
   $str =~ s/\&quot\;/\"/gi;
   $str =~ s/\&#47\;/\//gi;
   $str =~ s/\&#92\;/\\/gi;

   return $str;
}
